unit SDUComCtrls;

// TSDListView: An extended TListView
// Inherits from TListView, but includes functionality when in vsReport mode
// such that events are available when columns are resized, and supports
// ordering columns  

interface

uses
  ComCtrls, Classes, Messages, windows, CommCtrl;

type


  PNMHEADER = ^TNMHEADER;
  {$EXTERNALSYM tagNMHEADER}
  tagNMHEADER = packed record
    hdr: NMHDR;
    iItem: integer;
    iButton: integer;
    pitem: PHdItem;//PHDITEM;
  end;
  TNMHEADER = tagNMHEADER;
  {$EXTERNALSYM NMHEADER}
  NMHEADER = tagNMHEADER;

  TColumnResizeEvent = procedure (Sender: TListview; ColumnIdx: integer; NewWidth: integer) of object;
  TColumnSortEvent = procedure (Sender: TListview; ColumnIdx: integer) of object;

  TSDListView = class(TListView)
  private
    FColumnToSort: integer;

    FOnBeforeColumnSort: TColumnSortEvent;
    FOnAfterColumnSort: TColumnSortEvent;

    FOnBeforeColumnResize: TColumnResizeEvent;
    FOnColumnResize: TColumnResizeEvent;
    FOnAfterColumnResize: TColumnResizeEvent;

    procedure DoOnBeforeColumnSort(ColumnIdx: integer);
    procedure DoOnAfterColumnSort(ColumnIdx: integer);

    procedure DoOnBeforeColumnResize(ColumnIdx: integer; newWidth: integer);
    procedure DoOnColumnResize(ColumnIdx: integer; newWidth: integer);
    procedure DoOnAfterColumnResize(ColumnIdx: integer; newWidth: integer);

    procedure DoColumnSort(Column: TListColumn);

    procedure WMNotify(var Message: TWMNotify); message WM_NOTIFY; 

  protected
    procedure ColClick(Column: TListColumn); override;

    // specify -1 (HEADER_ROW) as y to get column headers
    function  GetCellText(x, y: integer): string;
    procedure SetCellText(x, y: integer; newText: string);

    function  GetSelectedIdx(): integer;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy(); override;

    function  AddCol(colCaption: string): TListColumn;
    function  InsertRow(position: integer; rowFields: TStringList; data: Pointer = nil): TListItem; overload;
    function  InsertRow(position: integer; rowFields: array of const; data: Pointer = nil): TListItem; overload;
    function  AppendRow(rowFields: TStringList; data: Pointer = nil): TListItem; overload;
    function  AppendRow(rowFields: array of const; data: Pointer = nil): TListItem; overload;

    procedure ResizeColumns();
    procedure _ColumnSortCompare(Sender: TObject; Item1, Item2: TListItem; Data: Integer; var Compare: Integer); virtual;

    procedure SwapItems(a: integer; b: integer);

    procedure SelectedMoveUp();
    procedure SelectedMoveDown();
    procedure SelectedDelete();
    function  GetFocused(): integer;
    property  CellText[x: integer; y: integer]: string read GetCellText write SetCellText;
  published
    property OnBeforeColumnSort: TColumnSortEvent read FOnBeforeColumnSort write FOnBeforeColumnSort;
    property OnAfterColumnSort: TColumnSortEvent read FOnAfterColumnSort write FOnAfterColumnSort;
    
    property OnBeforeColumnResize: TColumnResizeEvent read FOnBeforeColumnResize write FOnBeforeColumnResize;
    property OnColumnResize: TColumnResizeEvent read FOnColumnResize write FOnColumnResize;
    property OnAfterColumnResize: TColumnResizeEvent read FOnAfterColumnResize write FOnAfterColumnResize;

    // Note: This returns the *lowest* index number of all sected items.
    // Returns -1 if none selected
    property SelectedIdx: integer read GetSelectedIdx;

  end;

procedure Register;

implementation

uses
  SysUtils, 
  SDUGeneral;

const
  HEADER_ROW = -1;

procedure Register;
begin
  RegisterComponents('SDeanUtils', [TSDListView]);
end;

constructor TSDListView.Create(AOwner: TComponent);
begin
  inherited;

  ViewStyle := vsReport;
  
  FColumnToSort := 0;
end;

destructor TSDListView.Destroy();
begin
  inherited;
end;

procedure TSDListView.ColClick(Column: TListColumn);
begin
  DoOnBeforeColumnSort(Column.Index);
  DoColumnSort(Column);
  inherited;
  DoOnAfterColumnSort(Column.Index);
end;

function DefaultListViewColumnSort(Item1, Item2: TListItem;
  lParam: Integer): Integer; stdcall;
begin
  if Assigned(TListView(Item1.ListView).OnCompare) then
  begin
    TListView(Item1.ListView).OnCompare(Item1.ListView, Item1, Item2, lParam, Result);
  end
  else
  begin
    TSDListView(Item1.ListView)._ColumnSortCompare(Item1.ListView, Item1, Item2, lParam, Result);
  end;
end;

procedure TSDListView.DoColumnSort(Column: TListColumn);
begin
  FColumnToSort := Column.Index;

  // Don't call AlphaSort(...) - that uses the default sort; instead do what
  // AlphaSort(...) does, but call *our* version
  if HandleAllocated then
    begin
    ListView_SortItems(Handle, @DefaultListViewColumnSort, 0);
    end;

  if (FColumnToSort = 0) then
    begin
    if (Tag = 1) then
      begin
      Tag := 0;
      end
    else
      begin
      Tag := 1;
      end;

    end
  else
    begin
    if (Columns.Items[FColumnToSort-1].Tag = 1) then
      begin
      Columns.Items[FColumnToSort-1].Tag := 0;
      end
    else
      begin
      Columns.Items[FColumnToSort-1].Tag := 1;
      end;

    end;

end;

procedure TSDListView._ColumnSortCompare(Sender: TObject; Item1,
  Item2: TListItem; Data: Integer; var Compare: Integer);

  function CompareNumbers(ext1, ext2: extended): integer;
  var
    retval: integer;
  begin
    retval := 0;

    if (ext1 > ext2) then
      begin
      retval := 1;
      end
    else if (ext1 < ext2) then
      begin
      retval := -1;
      end;

    Result := retval;
  end;

  var
  str1, str2: string;
  colTag: integer;
  ext1, ext2: Extended;
  areNumbers: boolean;
begin
  if (FColumnToSort = 0) then
    begin
    ColTag := Tag;
    Str1 := Item1.Caption;
    Str2 := Item2.Caption;
    end
  else
    begin
    ColTag := Columns.Items[FColumnToSort-1].Tag;
    Str1 := Item1.SubItems.Strings[FColumnToSort-1];
    Str2 := Item2.SubItems.Strings[FColumnToSort-1];
    end;

  try
    ext1 := StrToFloat(Str1);
    ext2 := StrToFloat(Str2);
    areNumbers := TRUE;

    if (ColTag = 1) then
      begin
      Compare := CompareNumbers(ext1, ext2);
      end
    else
      begin
      Compare := CompareNumbers(ext2, ext1);
      end;
  except
    areNumbers := FALSE;
  end;

  if not(areNumbers) then
    begin
    if (ColTag = 1) then
      begin
      Compare := CompareText(Str1, Str2);
      end
    else
      begin
      Compare := CompareText(Str2, Str1);
      end;
    end;

end;

function TSDListView.AddCol(colCaption: string): TListColumn;
var
  newCol: TListColumn;
begin
  newCol := Columns.Add();
  newCol.Caption := colCaption;
  // Note: DON'T set AutoSize to TRUE; otherwise if the user resizes the
  //       dialog, any user set column widths will be reverted
  // newCol.AutoSize := TRUE;
  Result := newCol;
end;

function TSDListView.InsertRow(position: integer; rowFields: array of const; data: Pointer = nil): TListItem;
var
  i: integer;
  stlArgs: TStringList;
begin
  stlArgs:= TStringList.Create();
  try
    for i:=low(rowFields) to high(rowFields) do
      begin
      case rowFields[i].vtype of
        vtInteger    : stlArgs.Add(inttostr(rowFields[i].vinteger));
        vtBoolean    : stlArgs.Add(SDUBoolToStr(rowFields[i].vboolean));
        vtChar       : stlArgs.Add(rowFields[i].vchar);
        vtExtended   : stlArgs.Add(FloatToStr(rowFields[i].VExtended^));
        vtString     : stlArgs.Add(rowFields[i].VString^);
        vtPointer    : stlArgs.Add('0x'+inttohex(integer(rowFields[i].VPointer), 8));
        vtPChar      : stlArgs.Add(rowFields[i].VPChar);
        vtObject     : stlArgs.Add(rowFields[i].VObject.Classname);
        vtClass      : stlArgs.Add(rowFields[i].VClass.Classname);
        vtWideChar   : stlArgs.Add(rowFields[i].VWideChar);
        vtPWideChar  : stlArgs.Add(rowFields[i].VPWideChar);
        vtAnsiString : stlArgs.Add(PChar(rowFields[i].VAnsiString));
//        vtCurrency   : stlArgs.Add(rowFields[i].VCurrency);
        vtVariant    : stlArgs.Add(rowFields[i].VVariant^);
//        vtInterface  : stlArgs.Add(rowFields[i].VInterface);
        vtWideString : stlArgs.Add(PWideChar(rowFields[i].VWideString));
        vtInt64      : stlArgs.Add(inttostr(int64(rowFields[i].VInt64)));

      else
        stlArgs.Add('0x'+inttohex(rowFields[i].vtype, 2));
      end;

      end;

    Result := InsertRow(position, stlArgs, data);
    
  finally
    stlArgs.Free();
  end;

end;

function TSDListView.InsertRow(position: integer; rowFields: TStringList; data: Pointer = nil): TListItem;
var
  item: TListItem;
  stlPaddedFields: TStringList;
  i: integer;
begin
  stlPaddedFields:= TStringList.Create();
  try
    stlPaddedFields.Assign(rowFields);

    // Pad with blank strings so we have enough for each column
    for i:=stlPaddedFields.count to Columns.count do
      begin
      stlPaddedFields.Add('');
      end;

    item := Items.Insert(position);
    item.data := data;
    item.Caption := stlPaddedFields[0];

    // Start from 1 here; we've already done the initial caption
    for i:=1 to (stlPaddedFields.count - 1) do
      begin
      item.SubItems.Add(stlPaddedFields[i]);
      end;

  finally
    stlPaddedFields.Free();
  end;

  Result := item;
end;

function TSDListView.AppendRow(rowFields: TStringList; data: Pointer = nil): TListItem;
begin
  Result := InsertRow(Items.count, rowFields, data);
end;

function TSDListView.AppendRow(rowFields: array of const; data: Pointer = nil): TListItem;
begin
  Result := InsertRow(Items.count, rowFields, data);
end;

procedure TSDListView.ResizeColumns();
const
  // Resize the columns such that they're as wide as the widest item/subitem
  // text
  RESIZE_EXCL_HEADER = -1;
  // Resize the columns such that they're as wide as the column header text/the
  // widest item/subitem
  RESIZE_INCL_HEADER   = -2;
var
  i: integer;
  prevAutoSize: boolean;
begin
  for i:=0 to (Columns.count -1) do
    begin
    prevAutoSize := Column[i].AutoSize;
    Column[i].AutoSize := TRUE;
    Column[i].width := RESIZE_INCL_HEADER;
    // Revert AutoSize...
    Column[i].AutoSize := prevAutoSize;
    end;
end;

function TSDListView.GetCellText(x, y: integer): string;
var
  itemText: string;
  item: TListItem;
begin
  if (y = HEADER_ROW) then
    begin
    itemText := Columns[x].caption
    end
  else
    begin
    item := Items[y];

    if (x = 0) then
      begin
      itemText := item.caption;
      end
    else
      begin
      itemText := item.SubItems.Strings[x-1];
      end;

    end;

  Result := itemText;
end;

procedure TSDListView.SetCellText(x, y: integer; newText: string);
var
  item: TListItem;
begin
  if (y = HEADER_ROW) then
    begin
    Columns[x].caption := newText;
    end
  else
    begin
    item := Items[y];

    if (x = 0) then
      begin
      item.caption := newText;
      end
    else
      begin
      item.SubItems.Strings[x-1] := newText;
      end;

    end;

end;

procedure TSDListView.SwapItems(a: integer; b: integer);
var
  tempLI: TListItem;
begin
  Items.BeginUpdate;
  try
    tempLI := TListItem.Create(Items);
    tempLI.Assign(Items.Item[a]);
    Items.Item[a].Assign(Items.Item[b]);
    Items.Item[b].Assign(tempLI);
    tempLI.Free;
  finally
    Items.EndUpdate
  end;
end;


function TSDListView.GetFocused(): integer;
var
  i: integer;
  retval: integer;
begin
  retval := -1;

  for i:=0 to (items.count-1) do
    begin
    if Items[i].focused then
      begin
      retval := i;
      break;
      end;
    end;

  Result := retval;
end;

procedure TSDListView.SelectedMoveUp();
var
  i: integer;
  focusedIdx: integer;
  selStateOther: boolean;
begin
  focusedIdx:= GetFocused();

  // Start from 1 - can't move the first item up
  for i:=1 to (items.count-1) do
    begin
    if items[i].selected then
      begin
      selStateOther:= items[i-1].selected;

      SwapItems(i, (i-1));

      items[i].selected := selStateOther;
      items[i-1].selected := TRUE;
      end;
    end;


  if (
      ((focusedIdx-1) >= 0) and
      ((focusedIdx-1) <= (Items.count - 1))
     ) then
  begin
    items[focusedIdx-1].focused := TRUE;
  end;

end;

procedure TSDListView.SelectedMoveDown();
var
  i: integer;
  focusedIdx: integer;
  selStateOther: boolean;
begin
  focusedIdx:= GetFocused();

  // -2 because we can't move the last item down
  for i:=(items.count-2) downto 0 do
    begin
    if items[i].selected then
      begin
      selStateOther:= items[i+1].selected;

      SwapItems(i, (i+1));

      items[i].selected := selStateOther;
      items[i+1].selected := TRUE;
      end;
    end;

  if (
      ((focusedIdx+1) >= 0) and
      ((focusedIdx+1) <= (Items.count - 1))
     ) then
  begin
    items[focusedIdx+1].focused := TRUE;
  end;

end;

procedure TSDListView.SelectedDelete();
var
  i: integer;
begin
  for i:=(items.count-1) downto 0 do
    begin
    if items[i].selected then
      begin
      Items[i].Delete;
      end;
    end;

end;

procedure TSDListView.WMNotify(var Message: TWMNotify);
var
  header: PNMHEADER;
  newWidth: integer;
begin

  case Message.NMHdr.code of
//    HDN_BEGINTRACK:
    HDN_BEGINTRACKA,
    HDN_BEGINTRACKW,
//    HDN_TRACK:
    HDN_TRACKA,
    HDN_TRACKW,
//    HDN_ENDTRACK,
    HDN_ENDTRACKA,
    HDN_ENDTRACKW:
      begin
        newWidth := -1;
        header := PNMHEADER(Message.NMHdr);
        if (header.pitem <> nil) then
          begin
          if ((header.pItem.mask and HDI_WIDTH) <> 0) then
            begin
            newWidth := header.pItem.cxy;
            end;
          end;


      case Message.NMHdr.code of
    //    HDN_BEGINTRACK:
        HDN_BEGINTRACKA,
        HDN_BEGINTRACKW:
          begin
          DoOnBeforeColumnResize(header.iItem, newWidth);
          end;

    //    HDN_TRACK:
        HDN_TRACKA,
        HDN_TRACKW:
          begin
          DoOnColumnResize(header.iItem, newWidth);
          end;

    //    HDN_ENDTRACK,
        HDN_ENDTRACKA,
        HDN_ENDTRACKW:
          begin
          DoOnAfterColumnResize(header.iItem, newWidth);
          end;

        end;
      end;

    end;

  inherited;

end;

procedure TSDListView.DoOnBeforeColumnSort(ColumnIdx: integer);
begin
  if assigned(FOnBeforeColumnSort) then
    begin
    FOnBeforeColumnSort(self, ColumnIdx);
    end;
end;

procedure TSDListView.DoOnAfterColumnSort(ColumnIdx: integer);
begin
  if assigned(FOnAfterColumnSort) then
    begin
    FOnAfterColumnSort(self, ColumnIdx);
    end;
end;

procedure TSDListView.DoOnBeforeColumnResize(ColumnIdx: integer; newWidth: integer);
begin
  if assigned(FOnBeforeColumnResize) then
    begin
    FOnBeforeColumnResize(self, ColumnIdx, newWidth);
    end;
end;

procedure TSDListView.DoOnColumnResize(ColumnIdx: integer; newWidth: integer);
begin
  if assigned(FOnColumnResize) then
    begin
    FOnColumnResize(self, ColumnIdx, newWidth);
    end;
end;

procedure TSDListView.DoOnAfterColumnResize(ColumnIdx: integer; newWidth: integer);
begin
  if assigned(FOnAfterColumnResize) then
    begin
    FOnAfterColumnResize(self, ColumnIdx, newWidth);
    end;
end;

function TSDListView.GetSelectedIdx(): integer;
var
  retval: integer;
  i: integer;
begin
  retval := -1;

  for i:=0 to (self.Items.count - 1) do
    begin
    if self.Items[i].Selected then
      begin
      retval := i;
      break;
      end;
    end;

  Result := retval;
end;

END.

